@@ -1,4 +1,5 @@
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
 
 -- | In-memory TLS session manager.
 --
@@ -13,12 +14,18 @@ module Network.TLS.SessionManager (
   , newSessionManager
   ) where
 
+import Basement.Block (Block)
+import Data.ByteArray (convert)
 import Control.Exception (assert)
 import Control.Reaper
+import Data.ByteString (ByteString)
 import Data.IORef
 import Data.OrdPSQ (OrdPSQ)
 import qualified Data.OrdPSQ as Q
-import Network.TLS (SessionID, SessionData, SessionManager(..))
+import Network.TLS
+#if !MIN_VERSION_tls(1,5,0)
+import Network.TLS.Compression
+#endif
 import qualified System.Clock as C
 
 import Network.TLS.Imports
@@ -35,20 +42,68 @@ data Config = Config {
     , dbMaxSize      :: !Int
     }
 
--- | Lifetime: 1 day , delay: 10 minutes, limit: 10,000 entries.
+-- | Lifetime: 1 day , delay: 10 minutes, max size: 1000 entries.
 defaultConfig :: Config
 defaultConfig = Config {
       ticketLifetime = 86400
     , pruningDelay   = 6000
-    , dbMaxSize      = 10000
+    , dbMaxSize      = 1000
     }
 
 ----------------------------------------------------------------
 
+toKey :: ByteString -> Block Word8
+toKey = convert
+
+toValue :: SessionData -> SessionDataCopy
+#if MIN_VERSION_tls(1,5,0)
+toValue (SessionData v cid comp msni sec mg mti malpn siz) =
+    SessionDataCopy v cid comp msni sec' mg mti malpn' siz
+  where
+    !sec' = convert sec
+    !malpn' = convert <$> malpn
+#else
+toValue (SessionData v cid comp msni sec) =
+    SessionDataCopy v cid comp msni sec'
+  where
+    !sec' = convert sec
+#endif
+
+fromValue :: SessionDataCopy -> SessionData
+#if MIN_VERSION_tls(1,5,0)
+fromValue (SessionDataCopy v cid comp msni sec' mg mti malpn' siz) =
+    (SessionData v cid comp msni sec mg mti malpn siz)
+  where
+    !sec = convert sec'
+    !malpn = convert <$> malpn'
+#else
+fromValue (SessionDataCopy v cid comp msni sec') =
+    (SessionData v cid comp msni sec)
+  where
+    !sec = convert sec'
+#endif
+
+----------------------------------------------------------------
+
+type SessionIDCopy = Block Word8
+data SessionDataCopy = SessionDataCopy {
+      ssVersion     :: !Version
+    , ssCipher      :: !CipherID
+    , ssCompression :: !CompressionID
+    , ssClientSNI   :: !(Maybe HostName)
+    , ssSecret      :: Block Word8
+#if MIN_VERSION_tls(1,5,0)
+    , ssGroup       :: !(Maybe Group)
+    , ssTicketInfo  :: !(Maybe TLS13TicketInfo)
+    , ssALPN        :: !(Maybe (Block Word8))
+    , ssMaxEarlyDataSize :: Int
+#endif
+    } deriving (Show,Eq)
+
 type Sec = Int64
-type Value = (SessionData, IORef Availability)
-type DB = OrdPSQ SessionID Sec Value
-type Item = (SessionID, Sec, Value, Operation)
+type Value = (SessionDataCopy, IORef Availability)
+type DB = OrdPSQ SessionIDCopy Sec Value
+type Item = (SessionIDCopy, Sec, Value, Operation)
 
 data Operation = Add | Del
 data Use = SingleUse | MultipleUse
@@ -69,9 +124,13 @@ newSessionManager conf = do
         , reaperDelay  = pruningDelay conf * 1000000
         }
     return SessionManager {
-        sessionResume     = resume reaper MultipleUse
-      , sessionEstablish  = establish reaper lifetime
-      , sessionInvalidate = invalidate reaper
+        sessionResume         = resume reaper MultipleUse
+#if MIN_VERSION_tls(1,5,0)
+      , sessionResumeOnlyOnce = resume reaper SingleUse
+#endif
+      , sessionEstablish      = establish reaper lifetime
+      , sessionInvalidate     = invalidate reaper
+
       }
 
 cons :: Int -> Item -> DB -> DB
@@ -103,30 +162,36 @@ establish :: Reaper DB Item -> Sec
 establish reaper lifetime k sd = do
     ref <- newIORef Fresh
     !p <- ((+ lifetime) . C.sec) <$> C.getTime C.Monotonic
-    let !v = (sd,ref)
-    reaperAdd reaper (k,p,v,Add)
+    let !v = (sd',ref)
+    reaperAdd reaper (k',p,v,Add)
+  where
+    !k' = toKey k
+    !sd' = toValue sd
 
 resume :: Reaper DB Item -> Use
        -> SessionID -> IO (Maybe SessionData)
 resume reaper use k = do
     db <- reaperRead reaper
-    case Q.lookup k db of
+    case Q.lookup k' db of
       Nothing             -> return Nothing
       Just (p,v@(sd,ref)) -> do
            case use of
                SingleUse -> do
                    available <- atomicModifyIORef' ref check
-                   reaperAdd reaper (k,p,v,Del)
-                   return $ if available then Just sd else Nothing
-               MultipleUse -> return $ Just sd
+                   reaperAdd reaper (k',p,v,Del)
+                   return $ if available then Just (fromValue sd) else Nothing
+               MultipleUse -> return $ Just (fromValue sd)
   where
     check Fresh = (Used,True)
     check Used  = (Used,False)
+    !k' = toKey k
 
 invalidate :: Reaper DB Item
            -> SessionID -> IO ()
 invalidate reaper k = do
     db <- reaperRead reaper
-    case Q.lookup k db of
+    case Q.lookup k' db of
       Nothing    -> return ()
-      Just (p,v) -> reaperAdd reaper (k,p,v,Del)
+      Just (p,v) -> reaperAdd reaper (k',p,v,Del)
+  where
+    !k' = toKey k
